"""Fixpoint iterator over a CfG

- Transfer is a base class that must be instantiated for
each static analysis.
- fixpoint_iteration computes a fixpoint over a cfg for the given
transfer instance
"""
from cfg import *
from typing import Protocol

class Transfer[S](Protocol):
    """lists the functions that must be implemented for a static analysis over Lattice S"""

    def widen(self, s1: S, s2: S) -> S:
        """widening operator used on back edges (default: join)"""
        return self.join(s1, s2)
    
    def bottom(self) -> S:
        """the empty state, associated to unvisited nodes"""
        ...
    def init_state(self) -> S:
        """the state associated to the initial node"""
        ...
    def join(self,s1:S,s2:S) -> S:
        """join of two states when merging branches"""
        ...
    def included(self,s1: S,s2: S) -> bool:
        """return True when the first state is included in the second"""
        ...
    def tr_skip(self,s: S) -> S:
        """transfer state across a Skip transition

        Almost always transfer the state unchanged
        """
        return s
    def tr_set(self,s: S,v: str,e: ArithExpr) -> S:
        """transfer state across an assignment"""
        ...
    def tr_test(self,s: S,c: BoolExpr) -> S:
        """transfer state across a test transition"""
        ...
    def tr_err(self,s: S,e: Expr) -> S:
        """transfer state across a transition to error state"""
        ...

# def fixpoint_iteration[T](transfer: Transfer[T], cfg: Cfg) -> dict[Node,T]:
#     """computes the fixpoint for the given transfer functions over the given CfG
 
#    output maps each node to the computed state
#     """

#     # Initialize the mapping
#     mapping = {}
#     for node in cfg.g.nodes:
#         if node == cfg.init_node:
#             mapping[node] = transfer.init_state()
#         else:
#             mapping[node] = transfer.bottom()
    
#     # Initialize the worklist with all edges from the initial node
#     worklist = []
#     for (dst, label) in cfg.init_node.succs:
#         worklist.append((cfg.init_node, dst, label))
    
#     # Main iteration loop
#     while worklist:
#         # Take the first edge from the worklist
#         src, dst, label = worklist.pop(0)
        
#         # Apply the transfer function
#         src_state = mapping[src]
#         new_state = transfer.transfer(label, src_state)
        
#         # Compute the join with the current state
#         result_state = transfer.join(mapping[dst], new_state)
        
#         # Check if we need to update and propagate
#         if not transfer.le(result_state, mapping[dst]):
#             mapping[dst] = result_state
            
#             # Add all outgoing edges from dst to the worklist
#             for (next_dst, next_label) in dst.succs:
#                 worklist.append((dst, next_dst, next_label))
    
#     return mapping


def fixpoint_iteration[T](transfer: Transfer[T], cfg: Cfg) -> dict[Node,T]:
    """computes the fixpoint for the given transfer functions over the given CfG

    output maps each node to the computed state
    """

    # initialisation des états
    mapping: dict[Node, T] = {}
    for node in cfg.g.nodes:
        if node == cfg.init_node:
            mapping[node] = transfer.init_state()
        else:
            mapping[node] = transfer.bottom()

    # initialisation de la worklist avc les successeurs du noeud initial
    worklist: list[tuple[Node, Node, Label]] = []
    for (dst, label) in cfg.init_node.succs:
        worklist.append((cfg.init_node, dst, label))

    # iter principale
    while worklist:
        src, dst, label = worklist.pop(0)
        src_state = mapping[src]

        # applique la fonction de transfert en fonction du label
        match label:
            case LSkip():
                new_state = transfer.tr_skip(src_state)
            case LSet(var, expr):
                new_state = transfer.tr_set(src_state, var, expr)
            case LTest(cond):
                new_state = transfer.tr_test(src_state, cond)
            case LErr(expr):
                new_state = transfer.tr_err(src_state, expr)
            case _:
                continue  # ou raise NotImplementedError

        # appliquer le widen sur un arc de retour, sinon un simple join
        if isinstance(label, LSkip) and label.is_back_edge:
            joined_state = transfer.widen(mapping[dst], new_state)
        else:
            joined_state = transfer.join(mapping[dst], new_state)

        # Màj si l'état évolue
        if not transfer.included(joined_state, mapping[dst]):
            mapping[dst] = joined_state
            for (succ, lab) in dst.succs:
                worklist.append((dst, succ, lab))

    return mapping
